# vasil@vasil.com

implement Wbmp;

include "sys.m";
include "draw.m";
include "tk.m";
include "wmlib.m";

sys: Sys;
draw: Draw;
tk: Tk;
wmlib: Wmlib;
stderr : ref Sys->FD;
toplevel : ref Tk->Toplevel;
debug := 0;

include "wbmp.m";
include "loadfile.b";
include "tkcmd.b";
include "multibyte.b";

init(ctxt: ref Draw->Context, argv: list of string)
{
	s, fname : string;

	modinit();

	stderr = sys->fildes(2);
	wmlib->init();
	(win, winctl) := wmlib->titlebar(ctxt.screen, "", "WBMP Viewer", Wmlib->Appl);
	toplevel = win;

	argv = tl argv;
	while(argv != nil)
	{
		case (hd argv)
		{
		"-d" =>	debug = 1;
		* => fname = hd argv;
		}
		argv = tl argv;
	}

	cmd := chan of string;
	tk->namechan(win, cmd, "user");

	tkcmd("bind . <Key> {send user k %s}");
	tkcmd("bind . <ButtonRelease-1> {focus .}");
	tkcmd("bind .Wm_t <ButtonRelease-1> +{focus .}");
	tkcmd("frame .f -relief raised -bd 2");
	display(fname);
	tkcmd("pack .f -fill both -expand 1");
	tkcmd("pack propagate . 0 ; update");

	while(1)
	alt
	{
	s = <- cmd =>
		{
			sys->print("%s\n", s);
		}
	s = <- winctl =>
		{
			if(debug)
				sys->print("winctl: %s\n", s);
			if(s == "exit")
				exit;
			else
				wmlib->titlectl(win, s);
		}
	}
}

modinit()
{
	sys		= load Sys		Sys->PATH;
	draw	= load Draw		Draw->PATH;
	tk		= load Tk		Tk->PATH;
	wmlib	= load Wmlib	Wmlib->PATH;
}

display(f : string)
{
	b := array of byte loadfile(openfile(f));
	widget := wbmp(b, toplevel, ".f", 0);
	s := "pack " + widget + " -fill both -expand 1 ; update";
	tk->cmd(toplevel, s);
}

wbmp(b : array of byte, top : ref Tk->Toplevel, where : string, id : int) : string
{
	s : string;

	(i, h, w) := header(b);
	what := where + ".im" + string id;
	s = sys->sprint("canvas %s -height %d -width %d", what, h, w);
	tk->cmd(top, s);	

	# TODO: the tk channel name ("user" in this case) should be general.
	#
	s = sys->sprint("bind %s <ButtonRelease-1> {send user click %s}", what, what);
	tk->cmd(top, s);
	s = sys->sprint("scrollbar %s.scrolly -orient vertical -command {%s yview}", what, what);
	tk->cmd(top, s);
	s = sys->sprint("pack %s.scrolly -side right -fill y", what);
	tk->cmd(top, s);
	decode(b[i:], top, what, h, w);
	return what;
}

header(b : array of byte) : (int, int, int)
{
	w, h, n : int;

	i := 0;
	i++;	# type

#while((b[i] & byte 16r80) != byte 0)i++;
i += 1; # TODO: confirm this is correct. can someone find me a real wbmp spec?
	(n, w) = multibytetoint(b[i:]);
	i += n;
	(n, h) = multibytetoint(b[i:]);
	i += n;

	return (i, h, w);
}

decode(b : array of byte, top : ref Tk->Toplevel, f : string, hh, ww : int)
{
	#c : int;
	bb : byte;
	s : string;
	i := 0;

loop:
	for(h := 0; h < hh; h++)
	{
		for(w := 0; w < ww; w++)
		{
			if((w % 8) == 0)
			{
				if(i >= len b)
					break loop;
				bb = b[i];
				i++;
			}
			#c = ' ';
			if((bb & byte 128) == byte 0)
			{
				s = sys->sprint("%s create line %d %d %d %d -width 1 -fill black", f, w, h, w, h);
				tk->cmd(top, s);
				#c = '@';
			}
			#sys->print("%c", c);
			bb = bb << 1;
		}
		#sys->print("\n");
	}
}
