# vasil@vasil.com

implement WmlLib;

include "sys.m";
include "draw.m";
include "bufio.m";
include "string.m";
include "wmlbro.m";
include "wml.m";

sys : Sys;
draw : Draw;
str : String;
bufio	: Bufio;
	Iobuf : import bufio;
debug := 0;

stringtable : array of byte;

include "dumptagtree.b";
include "tagnode.b";
include "multibyte.b";

include "buildbinary.b";
include "buildtext.b";
include "parsebinary.b";

modinit()
{
	sys		= load Sys		Sys->PATH;
	bufio	= load Bufio	Bufio->PATH;
	str		= load String	String->PATH;
}

parsetext(s : string) : ref TagNode
{
	(nil, t) := parsewmltag(tokenize(removecomments(s)));
	return t;
}

removecomments(s : string) : string
{
	incomment := 0;
	closing : string;
	s2 : string;
	i := 0;
	j := 0;

	while(i < len s)
	{
		if(i+9 < len s && s[i:i+9] == "<!DOCTYPE")
		{
			i++;
			closing = ">";
			incomment = 1;
		}
		else if(i+2 < len s && s[i:i+2] == "<?")
		{
			i++;
			closing = "?>";
			incomment = 1;
		}
		else if(i+4 < len s && s[i:i+4] == "<!--")
		{
			i+=3;
			closing = "-->";
			incomment = 1;
		}
		if(incomment && s[i:i+len closing] == closing)
		{
			i+= len closing - 1;
			incomment = 0;
		}
		else if(!incomment)
		{
			s2[j] = s[i];
			j++;
		}
		i++;
	}
	return s2;
}

iswhite(c: int) : int
{
	return (c==' ' || c=='\t' || c=='\n' || c=='\r');
}

tokenize(s : string) : list of string
{
	i := 0;
	close := 0;
	intag := 0;
	word := "";
	if(s == nil)
		return nil;

	while(i < len s && iswhite(s[i]))
		i++;

	if(i < len s && s[i] == '<')
		intag = 1;

	while(i < len s)
	{
		if(close == 0)
		{
			if(intag && (s[i] == ''' || s[i] == '\"'))
				close = s[i];
			if(s[i] == '>')
			{
				word += s[i:i+1];
				i++;
				break;
			}
			else if(s[i] == '<')
			{
				if(word != nil)
					break;
				else
					word += s[i:i+1];
			}
			else
			{
				word += s[i:i+1];
			}
		}
		else
		{
			if(s[i] == close)
				close = 0;
			word += s[i:i+1];
		}
		i++;
	}
	if(word == nil)
		return tokenize(s[i:]);
	word = removetrailingws(word);
	return word :: tokenize(s[i:]);
}

removetrailingws(s : string) : string
{
	while(iswhite(s[len s - 1]))
	{
		s = s[:len s - 1];
	}
	return s;
}

parsewmltag(s : list of string) : (list of string, ref TagNode)
{
	newt : ref TagNode;
	empty := 0;
	t := ref TagNode;
	t.handle = nil;
	t.subtags = nil;
	w := hd s;

	# TODO: this is where we prevent HTML content from sneaking into the parser
	# note that if it is HTML, it will be retrned as a plain text literal and
	# displayed as such.
	# besides this, we should be checking the "content-type" field in that layer.

	if(w[0] == '<' && !str->prefix("html", w[1:]) && !str->prefix("HTML", w[1:]))
	{
		t.attribs = attachattribs(w);
		(n, ll) := sys->tokenize(w, " \t</>");
		justtag := hd ll;
		t.tag  = stringtotag(justtag);
		t.text = justtag;
		if(w[len w - 2] == '/')
		{
			empty = 1;
			s = tl s;
		}
		else
		{
			s = tl s;
loop1:
			while(s != nil)
			{
				# TODO: we should probably match the closing tag
				# with the tagname in the variable "justtag"
				#
				if(hassubstring(hd s, "</"))
				{
					s = tl s;
					break loop1;
				}
				else
				{
					(s, newt) = parsewmltag(s);
					t.subtags = addtoendoflist(t.subtags, newt);
				}
			}
		}
	}
	else
	{
		t.tag = LITERAL_C;
		t.text = w;
		s = tl s;
	}
	return (s, t);
}

hassubstring(s : string, ts : string) : int
{
	i := 0;
	while(i < len s)
	{
		if((len s[i:] >= len ts) && (s[i: i + len ts] == ts))
			return 1;
		i++;
	}
	return 0;
}

addtoendoflist(tlist : list of ref TagNode, t : ref TagNode) : list of ref TagNode
{
	if(tlist == nil)
	{
		return t :: nil;
	}
	return hd tlist :: addtoendoflist(tl tlist, t);
}

attachattribs(s : string) : list of ref TagNode
{
	attlist : list of ref TagNode;
	closing : int;

	attlist = nil;
	i := 0;
	end := 0;

	if(s[0] == '<')
		s = s[1:];
	if(s[len s - 2:len s] == "/>")
		s = s[0:len s - 2];
	if(s[len s - 1:len s] == ">")
		s = s[0:len s - 1];

	while(i < len s && !iswhite(s[i]))
		i++;
	if(i == len s)
		return nil;	# no attributes
	while(i < len s && iswhite(s[i]))
		i++;
	while(i < len s)
	{
		end = i;
		while(end < len s && s[end] != '\"' && s[end] != '\'')
		{
			if(s[end] == '=')
				s[end] = ' ';
			end++;
		}
		closing = s[end];
		end++;
		while(end < len s && s[end] != closing)
			end++;
		end++;
		t := attributepair(s[i:end]);
		i = end;
		attlist = addtoendoflist(attlist, t);
		while(i < len s && iswhite(s[i]))
			i++;
	}
	return attlist;
}

inquotes(s : string) : string
{
	end := -1;
	close := -1;
	i := 0;
loop1:
	while(i < len s)
	{
		if(s[i] == '\'' || s[i] == '\"')
		{
			close = i;
			break loop1;
		}
		i++;
	}
	if(close == -1)
		return nil;
	i++;
loop2:
	while(i < len s)
	{
		if(s[i] == s[close])
		{
			end = i;
			break loop2;
		}
		i++;
	}
	if(s[end] == s[close])
		return s[close+1:end];
	return nil;
}

attributepair(s : string) : ref TagNode
{
	(n, ll) := sys->tokenize(s, " \t\"\'");
	t := ref TagNode;
	t.handle = nil;
	t.tag	= stringtoattr(ll);
	t.text  = inquotes(s);
	t.attribs = nil;
	t.subtags = nil;
	return t;
}

stringtotag(word : string) : byte
{
	case word
	{
	"a" =>				return TAG_A;
	"anchor" =>			return TAG_ANCHOR;
	"access" =>			return TAG_ACCESS;
	"b" =>				return TAG_B;
	"big" =>			return TAG_BIG;
	"br" =>				return TAG_BR;
	"card" =>			return TAG_CARD;
	"do" =>				return TAG_DO;
	"em" =>				return TAG_EM;
	"fieldset" =>		return TAG_FIELDSET;
	"go" =>				return TAG_GO;
	"head" =>			return TAG_HEAD;
	"i" =>				return TAG_I;
	"img" =>			return TAG_IMG;
	"input" =>			return TAG_INPUT;
	"meta" =>			return TAG_META;
	"noop" =>			return TAG_NOOP;
	"p" =>				return TAG_P;
	"postfield" =>		return TAG_POSTFIELD;
	"pre" =>			return TAG_PRE;
	"prev" =>			return TAG_PREV;
	"onevent" =>		return TAG_ONEVENT;
	"optgroup" =>		return TAG_OPTGROUP;
	"option" =>			return TAG_OPTION;
	"refresh" =>		return TAG_REFRESH;
	"select" =>			return TAG_SELECT;
	"setvar" =>			return TAG_SETVAR;
	"small" =>			return TAG_SMALL;
	"strong" =>			return TAG_STRONG;
	"table" =>			return TAG_TABLE;
	"td" =>				return TAG_TD;
	"template" =>		return TAG_TEMPLATE;
	"timer" =>			return TAG_TIMER;
	"tr" =>				return TAG_TR;
	"u" =>				return TAG_U;
	"wml" =>			return TAG_WML;
	}
	return byte 1;	# what should unknown tag name return?
}

stringtoattr(ll : list of string) : byte
{
	word := hd ll;
	ll = tl ll;
	if(ll != nil)
		second := hd ll;
	case word
	{
	"accept-charset" =>	return ATT_ACCEPT_CHARSET;
	"accesskey" =>		return ATT_ACCESSKEY;
	"align" =>
						{
							case second
							{
							"top" =>	return ATT_ALIGN_TOP;
							"middle" =>	return ATT_ALIGN_MIDDLE;
							"bottom" =>	return ATT_ALIGN_BOTTOM;
							"L"or "left" =>	return ATT_ALIGN_LEFT;
							"C" or "center" =>	return ATT_ALIGN_CENTER;
							"R" or "right" =>	return ATT_ALIGN_RIGHT;
							* =>		return ATT_ALIGN;
							}		
						}
	"alt" =>			return ATT_ALT;
						#	return ATT_CACHE_CONTROL_NO_CACHE;
	"class" =>			return ATT_CLASS;
	"columns" =>		return ATT_COLUMNS;
	"content" =>		return ATT_CONTENT;
						#	return ATT_CONTENT_AVWW;
	"domain" =>			return ATT_DOMAIN;
	"emptyok" =>		if(second == "false")
							return ATT_EMPTYOK_FALSE;
						return ATT_EMPTYOK_TRUE;
	"enctype" =>		if(second == "application/x-www-form-urlencoded")
							return ATT_ENCTYPE_APP_X_W3_URLENCODED;
						return ATT_ENCTYPE;
						#	return ATT_ENCTYPE_MULTIPART_FORM_DATA;
	"format" =>			return ATT_FORMAT;
	"forua" =>			if(second == "false")
							return ATT_FORUA_FALSE;
						return ATT_FORUA_TRUE;
	"height" =>			return ATT_HEIGHT;
	"href" =>
						{
							if(str->prefix("http", second))
								return ATT_HREF_HTTP;
							if(str->prefix("https", second))
								return ATT_HREF_HTTPS;
							return ATT_HREF;
						}
	"hspace" =>			return ATT_HSPACE;
	"http-equiv" =>
						{
							case second
							{
							* =>		return ATT_HTTP_EQUIV;
							"expires" =>		return ATT_HTTP_EQUIV_EXPIRES;
							"content-type" =>	return ATT_HTTP_EQUIV_CONTENT_TYPE;
							}
						}
	"id" =>				return ATT_ID;
	"ivalue" =>			return ATT_IVALUE;
	"iname" =>			return ATT_INAME;
	"label" =>			return ATT_LABEL;
	"localsrc" =>		return ATT_LOCALSRC;
	"maxlength" =>		return ATT_MAXLENGTH;
	"method" =>
						{
							case second
							{
							"get" =>	return ATT_METHOD_GET;
							"post" =>	return ATT_METHOD_POST;
							}
						}
	"mode" =>
						{
							case second
							{
							"wrap" =>	return ATT_MODE_WRAP;
							"nowrap" =>	return ATT_MODE_NOWRAP;
							}
						}
	"multiple" =>		if(second == "false")
							return ATT_MULTIPLE_FALSE;
						return ATT_MULTIPLE_TRUE;
	"name" =>			return ATT_NAME;
	"newcontext" =>		if(second == "false")
							return ATT_NEWCONTEXT_FALSE;
						return ATT_NEWCONTEXT_TRUE;
	"onenterbackward" =>return ATT_ONENTERBACKWARD;
	"onenterforward" =>	return ATT_ONENTERFORWARD;
	"onpick" =>			return ATT_ONPICK;
	"ontimer" =>		return ATT_ONTIMER;
	"optional" =>		if(second == "false")
							return ATT_OPTIONAL_FALSE;
						return ATT_OPTIONAL_TRUE;
	"path" =>			return ATT_PATH;
	"scheme" =>			return ATT_SCHEME;
	"sendreferer" =>	if(second == "false")
							return ATT_SENDREFERER_FALSE;
						return ATT_SENDREFERER_TRUE;
	"size" =>			return ATT_SIZE;
	"src" =>
						{
							if(str->prefix("http", second))
								return ATT_SRC_HTTP;
							if(str->prefix("https", second))
								return ATT_SRC_HTTPS;
							return ATT_SRC;
						}
	"ordered" =>		if(second == "false")
							return ATT_ORDERED_FALSE;
						return ATT_ORDERED_TRUE;
	"tabindex" =>		return ATT_TABINDEX;
	"title" =>			return ATT_TITLE;
	"type" =>			
						{
							case second
							{
							"accept" =>	return ATT_TYPE_ACCEPT;
							"delete" =>	return ATT_TYPE_DELETE;
							"help" =>	return ATT_TYPE_HELP;
							"password" =>	return ATT_TYPE_PASSWORD;
							"onpick" =>	return ATT_TYPE_ONPICK;
							"onenterforward" =>	return ATT_TYPE_ONENTERFORWARD;
							"onenterbackward" =>return ATT_TYPE_ONENTERBACKWARD;
							"ontimer" =>	return ATT_TYPE_ONTIMER;
							"options" =>	return ATT_TYPE_OPTIONS;
							"prev" =>	return ATT_TYPE_PREV;
							"reset" =>	return ATT_TYPE_RESET;
							"text" =>	return ATT_TYPE_TEXT;
							#	"vnd" =>	return ATT_TYPE_VND;
							* =>		return ATT_TYPE;
							}		
						}
	"value" =>			return ATT_VALUE;
	"vspace" =>			return ATT_VSPACE;
	"width" =>			return ATT_WIDTH;
	"xml:lang" =>		return ATT_XML_LANG;
	"xml:space" =>		return ATT_XML_SPACE_PRESERVE;
						#	 return ATT_XML_SPACE_DEFAULT;
	}
	return byte 1;	# what should unknown tag name return?
}
